#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/miscdevice.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/seq_file.h>
#include <linux/spinlock.h>
#include <linux/sysctl.h>
#include <linux/wait.h>
#include <linux/bcd.h>
#include <linux/delay.h>
#include <linux/platform_device.h>

#include <mach/hardware.h>
#include <mach/regs-gpio.h>

#define LED_MINOR	249
#define LED_OFF		0
#define LED_DAY		1	
#define LED_NIGHT	2

static int led_open_cnt = 0;
static unsigned char led_level = 2;

static void led_set_level(unsigned char level)
{
	switch(level) {
	case LED_OFF:
		s3c2410_gpio_setpin(S3C2410_GPG6, 0);
		s3c2410_gpio_setpin(S3C2410_GPG7, 0);
		break;
	case LED_NIGHT:
		s3c2410_gpio_setpin(S3C2410_GPG6, 1);
		s3c2410_gpio_setpin(S3C2410_GPG7, 0);
		break;
	case LED_DAY:
		s3c2410_gpio_setpin(S3C2410_GPG6, 0); 
		s3c2410_gpio_setpin(S3C2410_GPG7, 1);
		break;
	}
}

static ssize_t
led_write(struct file *file, const char __user *buf, size_t count, loff_t *ppos)
{
	unsigned char level;

	if (copy_from_user(&level, buf, 1))
		return -EFAULT;

	if (level > 2)
		level = 2;
	else if (level <= 0)
		level = 0;
	led_level = level;
	led_set_level(level);

	return 1;
}

static int
led_open(struct inode *inode, struct file *file)
{
	if (led_open_cnt)
		return -EBUSY;
	led_open_cnt++;
	return 0;
}

static int
led_release(struct inode *inode, struct file *file)
{
	led_open_cnt--;
	return 0;
}

static const struct file_operations led_fops = {
	.owner		= THIS_MODULE,
	.write		= led_write,
	.open		= led_open,
	.release	= led_release,
};

static struct miscdevice led_dev = {
	LED_MINOR,
	"yup110n_led",
	&led_fops
};

static int __init led_probe(struct platform_device *pdev)
{
	int ret;

	s3c2410_gpio_cfgpin(S3C2410_GPG6, S3C2410_GPIO_OUTPUT);
        s3c2410_gpio_pullup(S3C2410_GPG6, S3C24XX_GPIO_PULL_DOWN);
	s3c2410_gpio_cfgpin(S3C2410_GPG7, S3C2410_GPIO_OUTPUT);
	s3c2410_gpio_pullup(S3C2410_GPG7, S3C24XX_GPIO_PULL_DOWN);

	s3c2410_gpio_setpin(S3C2410_GPG6, 1);
	s3c2410_gpio_setpin(S3C2410_GPG7, 0);
		
	ret = misc_register(&led_dev);
	if (ret) {
		printk(KERN_ERR "led: can't misc_register on minor=%d\n",
		       LED_MINOR);
		goto out;
	}
	ret = 0;
	printk(KERN_INFO "LED brightness control driver\n");
      out:
	return ret;
}

static int __exit led_remove(struct platform_device *pdev)
{
	misc_deregister(&led_dev);
	return 0;
}

#ifdef CONFIG_PM
static int led_suspend(struct platform_device *pdev, pm_message_t state)
{
	led_set_level(0);
	return 0;
}

static int led_resume(struct platform_device *pdev)
{
	led_set_level(led_level);
	return 0;
}
#else
#define led_suspend NULL
#define led_resume  NULL
#endif

static struct platform_driver led_driver = {
	.probe		= led_probe,
	.remove		= __exit_p(led_remove),
	.suspend	= led_suspend,
	.resume		= led_resume,
	.driver		= {
		.name	= "yup110n_led",
		.owner	= THIS_MODULE,
	},
};

static int __init led_init(void)
{
	return platform_driver_register(&led_driver);
}

static void __exit led_exit(void)
{
	platform_driver_unregister(&led_driver);
}

module_init(led_init);
module_exit(led_exit);

MODULE_LICENSE("GPL");
